@@ -4,14 +4,15 @@
 from __future__ import annotations
 
 __author__ = "ChenyangGao <https://chenyanggao.github.io>"
-__version__ = (0, 0, 3)
+__version__ = (0, 0, 4)
 __all__ = ["HTTPFileReader", "AsyncHTTPFileReader"]
 
 import errno
 
 from collections.abc import Awaitable, Callable, Mapping
 from functools import cached_property, partial
-from inspect import isawaitable, signature
+from importlib.util import find_spec
+from inspect import isawaitable
 from io import (
     BufferedReader, RawIOBase, TextIOWrapper, UnsupportedOperation, DEFAULT_BUFFER_SIZE, 
 )
@@ -19,7 +20,7 @@ from os import fstat, stat, PathLike
 from shutil import COPY_BUFSIZE # type: ignore
 from sys import exc_info
 from typing import cast, overload, Any, BinaryIO, Literal, Self
-from types import MappingProxyType, MethodType
+from types import MappingProxyType
 from warnings import warn
 
 from asynctools import ensure_async, run_async
@@ -102,7 +103,7 @@ class HTTPFileReader(RawIOBase, BinaryIO):
         # NOTE: If the offset of the forward seek is not higher than this value, 
         #       it will be directly read and discarded, default to 1 MB
         seek_threshold: int = 1 << 20, 
-        urlopen: Callable = urlopen, 
+        urlopen = urlopen, 
     ):
         if headers:
             headers = {**headers, "Accept-Encoding": "identity"}
@@ -112,6 +113,8 @@ class HTTPFileReader(RawIOBase, BinaryIO):
             headers["Range"] = f"bytes={start}-"
         elif start < 0:
             headers["Range"] = f"bytes={start}"
+        if urlopen is None:
+            urlopen = globals()["urlopen"]
         if callable(url):
             geturl = url
             def url():
@@ -212,6 +215,8 @@ class HTTPFileReader(RawIOBase, BinaryIO):
         return self.file.fileno()
 
     def flush(self, /):
+        if self.closed:
+            raise ValueError("I/O operation on closed file.")
         return self.file.flush()
 
     def isatty(self, /) -> bool:
@@ -250,7 +255,16 @@ class HTTPFileReader(RawIOBase, BinaryIO):
             return 0
         if self.file_closed:
             self.reconnect()
-        size = self.file.readinto(buffer)
+        try:
+            readinto = self.file.readinto
+        except AttributeError:
+            read = self.file.read
+            def readinto(buffer, /):
+                data = read(len(buffer))
+                size = len(data)
+                buffer[:size] = data
+                return data
+        size = readinto(buffer)
         if size:
             self._add_start(size)
         return size
@@ -262,6 +276,7 @@ class HTTPFileReader(RawIOBase, BinaryIO):
             return b""
         if self.file_closed:
             self.reconnect()
+        # TODO: if no readline method, give a default impl
         if size is None or size < 0:
             data = self.file.readline()
         else:
@@ -277,6 +292,7 @@ class HTTPFileReader(RawIOBase, BinaryIO):
             return []
         if self.file_closed:
             self.reconnect()
+        # TODO: if no readlines method, give a default impl
         ls = self.file.readlines(hint)
         if ls:
             self._add_start(sum(map(len, ls)))
@@ -384,7 +400,7 @@ class HTTPFileReader(RawIOBase, BinaryIO):
         headers: None | Mapping = None, 
         start: int = 0, 
         seek_threshold: int = 1 << 20, 
-        urlopen: Callable = urlopen, 
+        urlopen = None, 
     ) -> Self | BufferedReader:
         ...
     @overload
@@ -402,7 +418,7 @@ class HTTPFileReader(RawIOBase, BinaryIO):
         headers: None | Mapping = None, 
         start: int = 0, 
         seek_threshold: int = 1 << 20, 
-        urlopen: Callable = urlopen, 
+        urlopen = None, 
     ) -> TextIOWrapper:
         ...
     @classmethod
@@ -419,7 +435,7 @@ class HTTPFileReader(RawIOBase, BinaryIO):
         headers: None | Mapping = None, 
         start: int = 0, 
         seek_threshold: int = 1 << 20, 
-        urlopen: Callable = urlopen, 
+        urlopen = None, 
     ) -> Self | BufferedReader | TextIOWrapper:
         file = cls(
             url=url, 
@@ -517,9 +533,10 @@ class AsyncHTTPFileReader(HTTPFileReader):
         headers: None | Mapping = None, 
         start: int = 0, 
         seek_threshold: int = 1 << 20, 
-        urlopen: Callable = urlopen, 
+        urlopen = None, 
     ):
-        run_async(self._init(
+        self.__dict__.update(closed=False, start=0)
+        self.__dict__["_initing"] = run_async(self.__ainit__(
             url=url, 
             headers=headers, 
             start=start, 
@@ -528,6 +545,9 @@ class AsyncHTTPFileReader(HTTPFileReader):
         ))
 
     async def __aenter__(self, /) -> Self:
+        if _initing := self.__dict__.get("_initing"):
+            await _initing
+            self.__dict__.pop("_initing", None)
         return self
 
     async def __aexit__(self, /, *exc_info):
@@ -542,7 +562,7 @@ class AsyncHTTPFileReader(HTTPFileReader):
         else:
             raise StopAsyncIteration
 
-    async def _init(
+    async def __ainit__(
         self, 
         /, 
         url: str | Callable[[], str] | Callable[[], Awaitable[str]], 
@@ -551,8 +571,6 @@ class AsyncHTTPFileReader(HTTPFileReader):
         seek_threshold: int = 1 << 20, 
         urlopen = None, 
     ):
-        if urlopen is None:
-            urlopen = signature(type(self).__init__).parameters["urlopen"].default
         if headers:
             headers = {**headers, "Accept-Encoding": "identity"}
         else:
@@ -561,6 +579,8 @@ class AsyncHTTPFileReader(HTTPFileReader):
             headers["Range"] = f"bytes={start}-"
         elif start < 0:
             headers["Range"] = f"bytes={start}"
+        if urlopen is None:
+            urlopen = globals()["urlopen"]
         if callable(url):
             geturl = url
             async def url() -> str:
@@ -588,12 +608,12 @@ class AsyncHTTPFileReader(HTTPFileReader):
             length = get_total_length(response) or 0, 
             chunked = is_chunked(response), 
             start = start, 
-            closed = False, 
             urlopen = urlopen, 
             headers = MappingProxyType(headers), 
             seek_threshold = max(seek_threshold, 0), 
             _seekable = is_range_request(response), 
         )
+        self.__dict__.pop("_initing", None)
 
     async def aclose(self, /):
         if not self.closed:
@@ -606,7 +626,10 @@ class AsyncHTTPFileReader(HTTPFileReader):
             try:
                 ret = self.response.aclose()
             except (AttributeError, TypeError):
-                ret = self.response.close()
+                try:
+                    ret = self.response.close()
+                except (AttributeError, TypeError):
+                    return
             if isawaitable(ret):
                 run_async(ret)
 
@@ -614,24 +637,36 @@ class AsyncHTTPFileReader(HTTPFileReader):
         try:
             ret = self.response.aclose()
         except (AttributeError, TypeError):
-            ret = self.response.close()
+            try:
+                ret = self.response.close()
+            except (AttributeError, TypeError):
+                return
         if isawaitable(ret):
             await ret
 
     async def flush(self, /):
+        if self.closed:
+            raise ValueError("I/O operation on closed file.")
+        if _initing := self.__dict__.get("_initing"):
+            await _initing
+            self.__dict__.pop("_initing", None)
         return await ensure_async(self.file.flush, threaded=True)()
 
     async def read(self, size: int = -1, /) -> bytes: # type: ignore
         if self.closed:
             raise ValueError("I/O operation on closed file.")
+        if _initing := self.__dict__.get("_initing"):
+            await _initing
+            self.__dict__.pop("_initing", None)
         if size == 0 or not self.chunked and self.tell() >= self.length:
             return b""
         if self.file_closed:
             await self.reconnect()
+        read = ensure_async(self.file.read, threaded=True)
         if size is None or size < 0:
-            data = await ensure_async(self.file.read, threaded=True)()
+            data = await read(-1)
         else:
-            data = await ensure_async(self.file.read, threaded=True)(size)
+            data = await read(size)
         if data:
             self._add_start(len(data))
         return data
@@ -639,11 +674,23 @@ class AsyncHTTPFileReader(HTTPFileReader):
     async def readinto(self, buffer, /) -> int: # type: ignore
         if self.closed:
             raise ValueError("I/O operation on closed file.")
+        if _initing := self.__dict__.get("_initing"):
+            await _initing
+            self.__dict__.pop("_initing", None)
         if not buffer or not self.chunked and self.tell() >= self.length:
             return 0
         if self.file_closed:
             await self.reconnect()
-        size = await ensure_async(self.file.readinto, threaded=True)(buffer)
+        try:
+            readinto = ensure_async(self.file.readinto, threaded=True)
+        except AttributeError:
+            read = ensure_async(self.file.read, threaded=True)
+            async def readinto(buffer, /):
+                data = await read(len(buffer))
+                size = len(data)
+                buffer[:size] = data
+                return size
+        size = await readinto(buffer)
         if size:
             self._add_start(size)
         return size
@@ -651,10 +698,14 @@ class AsyncHTTPFileReader(HTTPFileReader):
     async def readline(self, size: None | int = -1, /) -> bytes: # type: ignore
         if self.closed:
             raise ValueError("I/O operation on closed file.")
+        if _initing := self.__dict__.get("_initing"):
+            await _initing
+            self.__dict__.pop("_initing", None)
         if size == 0 or not self.chunked and self.tell() >= self.length:
             return b""
         if self.file_closed:
             await self.reconnect()
+        # TODO: if no readline method, give a default impl
         if size is None or size < 0:
             data = await ensure_async(self.file.readline, threaded=True)()
         else:
@@ -666,16 +717,23 @@ class AsyncHTTPFileReader(HTTPFileReader):
     async def readlines(self, hint: int = -1, /) -> list[bytes]: # type: ignore
         if self.closed:
             raise ValueError("I/O operation on closed file.")
+        if _initing := self.__dict__.get("_initing"):
+            await _initing
+            self.__dict__.pop("_initing", None)
         if not self.chunked and self.tell() >= self.length:
             return []
         if self.file_closed:
             await self.reconnect()
+        # TODO: if no readlines method, give a default impl
         ls = await ensure_async(self.file.readlines, threaded=True)(hint)
         if ls:
             self._add_start(sum(map(len, ls)))
         return ls
 
     async def reconnect(self, /, start: None | int = None) -> int: # type: ignore
+        if _initing := self.__dict__.get("_initing"):
+            await _initing
+            self.__dict__.pop("_initing", None)
         if not self._seekable:
             if start is None and self.tell() or start:
                 raise OSError(errno.EOPNOTSUPP, "Unsupport for reconnection of non-seekable streams.")
@@ -708,6 +766,9 @@ class AsyncHTTPFileReader(HTTPFileReader):
     async def seek(self, pos: int, whence: int = 0, /) -> int: # type: ignore
         if self.closed:
             raise ValueError("I/O operation on closed file.")
+        if _initing := self.__dict__.get("_initing"):
+            await _initing
+            self.__dict__.pop("_initing", None)
         if not self._seekable:
             raise OSError(errno.EINVAL, "not a seekable stream")
         if whence == 0:
@@ -749,7 +810,7 @@ class AsyncHTTPFileReader(HTTPFileReader):
         urlopen = None, 
     ) -> Self: # type: ignore
         self = cls.__new__(cls)
-        await self._init(
+        await self.__ainit__(
             url=url, 
             headers=headers, 
             start=start, 
@@ -773,7 +834,7 @@ class AsyncHTTPFileReader(HTTPFileReader):
         headers: None | Mapping = None, 
         start: int = 0, 
         seek_threshold: int = 1 << 20, 
-        urlopen: Callable = urlopen, 
+        urlopen = None, 
     ) -> Self | AsyncBufferedReader:
         ...
     @overload
@@ -791,7 +852,7 @@ class AsyncHTTPFileReader(HTTPFileReader):
         headers: None | Mapping = None, 
         start: int = 0, 
         seek_threshold: int = 1 << 20, 
-        urlopen: Callable = urlopen, 
+        urlopen = None, 
     ) -> AsyncTextIOWrapper:
         ...
     @classmethod
@@ -807,7 +868,7 @@ class AsyncHTTPFileReader(HTTPFileReader):
         headers: None | Mapping = None, 
         start: int = 0, 
         seek_threshold: int = 1 << 20, 
-        urlopen: Callable = urlopen, 
+        urlopen = None, 
     ) -> Self | AsyncBufferedReader | AsyncTextIOWrapper:
         file = await cls.new(
             url=url, 
@@ -895,10 +956,10 @@ class AsyncHTTPFileReader(HTTPFileReader):
             return buffer
 
 
-try:
+if find_spec("urllib3"):
     from urllib.error import HTTPError
-    from urllib3 import request as urllib3_request
-    from urllib3.poolmanager import PoolManager
+
+    _urllib3_urlopen = None
 
     class Urllib3FileReader(HTTPFileReader):
 
@@ -909,17 +970,28 @@ try:
             headers: None | Mapping = None, 
             start: int = 0, 
             seek_threshold: int = 1 << 20, 
-            urlopen = partial(urllib3_request, "GET"), # type: ignore
+            urlopen = None, 
         ):
+            global _urllib3_urlopen
+            from urllib3 import request
+            from urllib3.poolmanager import PoolManager
+
             if isinstance(urlopen, PoolManager):
                 urlopen = partial(urlopen.request, "GET", preload_content=False)
+            elif urlopen is None:
+                if _urllib3_urlopen is None:
+                    urlopen = _urllib3_urlopen = partial(cast(Callable, request), "GET", preload_content=False)
+                else:
+                    urlopen = _urllib3_urlopen
             else:
                 urlopen = partial(urlopen, preload_content=False)
+
             def urlopen_wrapper(url: str, headers: None | Mapping = headers):
                 resp = urlopen(url, headers=headers)
                 if resp.status >= 400:
                     raise HTTPError(resp.url, resp.status, resp.reason, resp.headers, resp)
                 return resp
+
             super().__init__(
                 url, 
                 headers=headers, 
@@ -929,15 +1001,10 @@ try:
             )
 
     __all__.append("Urllib3FileReader")
-except ImportError:
-    pass
 
 
-try:
-    from requests import request as requests_request, Session
-
-    if "__del__" not in Session.__dict__:
-        setattr(Session, "__del__", lambda self, /: self.close())
+if find_spec("requests"):
+    _requests_urlopen = None
 
     class RequestsFileReader(HTTPFileReader):
 
@@ -948,10 +1015,19 @@ try:
             headers: None | Mapping = None, 
             start: int = 0, 
             seek_threshold: int = 1 << 20, 
-            urlopen = partial(requests_request, "GET"), 
+            urlopen = None, 
         ):
+            global _requests_urlopen
+            from requests import request, Session
             if isinstance(urlopen, Session):
                 urlopen = partial(urlopen.request, "GET", stream=True)
+            elif urlopen is None:
+                if _requests_urlopen is None:
+                    if "__del__" not in Session.__dict__:
+                        setattr(Session, "__del__", lambda self, /: self.close())
+                    urlopen = _requests_urlopen = partial(request, "GET", stream=True)
+                else:
+                    urlopen = _requests_urlopen
             else:
                 urlopen = partial(urlopen, stream=True)
             def urlopen_wrapper(url: str, headers: None | Mapping = headers):
@@ -971,33 +1047,16 @@ try:
             return self.response.raw
 
     __all__.append("RequestsFileReader")
-except ImportError:
-    pass
 
 
-try:
-    from aiohttp import request as aiohttp_request, ClientSession
+if find_spec("aiohttp"):
+    from types import MethodType
+
+    _aiohttp_urlopen = None
 
     class AiohttpFileReader(AsyncHTTPFileReader):
 
-        def __init__(
-            self, 
-            /, 
-            url: str | Callable[[], str] | Callable[[], Awaitable[str]], 
-            headers: None | Mapping = None, 
-            start: int = 0, 
-            seek_threshold: int = 1 << 20, 
-            urlopen = partial(aiohttp_request, "GET"), 
-        ):
-            super().__init__(
-                url=url, 
-                headers=headers, 
-                start=start, 
-                seek_threshold=seek_threshold, 
-                urlopen=urlopen, 
-            )
-
-        async def _init(
+        async def __ainit__(
             self, 
             /, 
             url: str | Callable[[], str] | Callable[[], Awaitable[str]], 
@@ -1006,17 +1065,26 @@ try:
             seek_threshold: int = 1 << 20, 
             urlopen = None, 
         ):
-            if urlopen is None:
-                urlopen = signature(type(self).__init__).parameters["urlopen"].default
+            global _aiohttp_urlopen
+            from aiohttp import request, ClientSession
+
             close_session = True
             if isinstance(urlopen, ClientSession):
                 urlopen = urlopen.get
-                do_not_close_session = False
-            if close_session:
+                close_session = False
+            if urlopen is None:
+                if _aiohttp_urlopen is None:
+                    urlopen = _aiohttp_urlopen = partial(request, "GET")
+                else:
+                    urlopen = _aiohttp_urlopen
+            else:
                 func = urlopen
                 if isinstance(func, partial):
                     func = func.func
-                close_session = not (isinstance(func, MethodType) and isinstance(func.__self__, ClientSession))
+                close_session = not (
+                    isinstance(func, MethodType) and 
+                    isinstance(func.__self__, ClientSession)
+                )
             async def urlopen_wrapper(url: str, headers: None | Mapping = headers):
                 resp = await urlopen(url, headers=headers).__aenter__()
                 async def aclose():
@@ -1029,7 +1097,7 @@ try:
                 resp.aclose = aclose
                 resp.raise_for_status()
                 return resp
-            await super()._init(
+            await super().__ainit__(
                 url=url, 
                 headers=headers, 
                 start=start, 
@@ -1046,60 +1114,13 @@ try:
             return self.response.closed
 
     __all__.append("AiohttpFileReader")
-except ImportError:
-    pass
 
 
-try:
-    from contextlib import asynccontextmanager
+if find_spec("httpx"):
     from filewrap import bytes_iter_to_reader, bytes_iter_to_async_reader
-    from httpx import stream, Client, AsyncClient
 
-    if "__del__" not in Client.__dict__:
-        setattr(Client, "__del__", lambda self: self.close())
-    if "__del__" not in AsyncClient.__dict__:
-        setattr(AsyncClient, "__del__", lambda self: run_async(self.aclose()))
-
-    def async_stream(
-        method, 
-        url, 
-        params = None,
-        content = None,
-        data = None,
-        files = None,
-        json = None,
-        headers = None,
-        cookies = None,
-        auth = None,
-        proxy = None,
-        proxies = None,
-        timeout = 5.0,
-        follow_redirects = False,
-        verify = True,
-        cert = None,
-        trust_env = True,
-    ):
-        client = AsyncClient(
-            cookies=cookies,
-            proxy=proxy,
-            proxies=proxies,
-            cert=cert,
-            verify=verify,
-            timeout=timeout,
-            trust_env=trust_env,
-        )
-        return client.stream(
-            method=method,
-            url=url,
-            content=content,
-            data=data,
-            files=files,
-            json=json,
-            params=params,
-            headers=headers,
-            auth=auth,
-            follow_redirects=follow_redirects,
-        )
+    _httpx_urlopen = None
+    _httpx_urlopen_async = None
 
     class HttpxFileReader(HTTPFileReader):
 
@@ -1110,10 +1131,21 @@ try:
             headers: None | Mapping = None, 
             start: int = 0, 
             seek_threshold: int = 1 << 20, 
-            urlopen = partial(stream, "GET"), 
+            urlopen = None, 
         ):
+            global _httpx_urlopen
+            from httpx import stream, Client
+
             if isinstance(urlopen, Client):
                 urlopen = partial(urlopen.stream, "GET")
+            elif urlopen is None:
+                if _httpx_urlopen is None:
+                    if "__del__" not in Client.__dict__:
+                        setattr(Client, "__del__", lambda self: self.close())
+                    urlopen = _httpx_urlopen = partial(stream, "GET")
+                else:
+                    urlopen = _httpx_urlopen
+
             def urlopen_wrapper(url: str, headers: None | Mapping = headers):
                 context = urlopen(url, headers=headers)
                 resp = context.__enter__()
@@ -1122,6 +1154,7 @@ try:
                 file = bytes_iter_to_reader(resp.iter_raw())
                 self.__dict__["file"] = file
                 return resp
+
             super().__init__(
                 url, 
                 headers=headers, 
@@ -1136,24 +1169,7 @@ try:
 
     class AsyncHttpxFileReader(AsyncHTTPFileReader):
 
-        def __init__(
-            self, 
-            /, 
-            url: str | Callable[[], str] | Callable[[], Awaitable[str]], 
-            headers: None | Mapping = None, 
-            start: int = 0, 
-            seek_threshold: int = 1 << 20, 
-            urlopen = partial(async_stream, "GET"), 
-        ):
-            super().__init__(
-                url=url, 
-                headers=headers, 
-                start=start, 
-                seek_threshold=seek_threshold, 
-                urlopen=urlopen, 
-            )
-
-        async def _init(
+        async def __ainit__(
             self, 
             /, 
             url: str | Callable[[], str] | Callable[[], Awaitable[str]], 
@@ -1162,10 +1178,59 @@ try:
             seek_threshold: int = 1 << 20, 
             urlopen = None, 
         ):
-            if urlopen is None:
-                urlopen = signature(type(self).__init__).parameters["urlopen"].default
+            global _httpx_urlopen_async
+            from httpx import AsyncClient
+
             if isinstance(urlopen, AsyncClient):
                 urlopen = partial(urlopen.stream, "GET")
+            elif urlopen is None:
+                if _httpx_urlopen_async is None:
+                    if "__del__" not in AsyncClient.__dict__:
+                       setattr(AsyncClient, "__del__", lambda self: run_async(self.aclose()))
+                    def async_stream(
+                        method, 
+                        url, 
+                        params = None, 
+                        content = None, 
+                        data = None, 
+                        files = None, 
+                        json = None, 
+                        headers = None, 
+                        cookies = None, 
+                        auth = None, 
+                        proxy = None, 
+                        proxies = None, 
+                        timeout = 5.0, 
+                        follow_redirects = False, 
+                        verify = True, 
+                        cert = None, 
+                        trust_env = True, 
+                    ):
+                        client = AsyncClient(
+                            cookies=cookies,
+                            proxy=proxy,
+                            proxies=proxies,
+                            cert=cert,
+                            verify=verify,
+                            timeout=timeout,
+                            trust_env=trust_env,
+                        )
+                        return client.stream(
+                            method=method,
+                            url=url,
+                            content=content,
+                            data=data,
+                            files=files,
+                            json=json,
+                            params=params,
+                            headers=headers,
+                            auth=auth,
+                            follow_redirects=follow_redirects,
+                        )
+                    urlopen = _httpx_urlopen_async = partial(async_stream, "GET")
+                else:
+                    urlopen = _httpx_urlopen_async
+
             async def urlopen_wrapper(url: str, headers: None | Mapping = headers):
                 context = urlopen(url, headers=headers)
                 resp = await context.__aenter__()
@@ -1174,7 +1239,8 @@ try:
                 file = bytes_iter_to_async_reader(resp.aiter_raw())
                 self.__dict__["file"] = file
                 return resp
-            await super()._init(
+
+            await super().__ainit__(
                 url=url, 
                 headers=headers, 
                 start=start, 
@@ -1188,7 +1254,6 @@ try:
 
     __all__.append("HttpxFileReader")
     __all__.append("AsyncHttpxFileReader")
-except ImportError:
-    pass
 
+# TODO: 增加 blacksheep 的 HTTPFileReader
 # TODO: 设计实现一个 HTTPFileWriter，用于实现上传，关闭后视为上传完成
